#!/bin/python3
import serial
import ctypes as ct
import time
import os 
from typing import Tuple, Union
from enum import Enum


SYNC_BYTE = 0xC8
CRSF_HEADER_SIZE = 5

M0139_RX_INFO_LEN = 30		# ModalAI M0139 
BETAFPV_RX_INFO_LEN = 32	# BetaFPV 900RX
FRSKY_RX_INFO_LEN = 30		# FrSky R9mm mini 

# Detect if we are on a voxl2 or host computer
if os.path.exists('/usr/share/modalai/voxl-elrs'):
    import voxl_serial
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "voxl2"
else:
    BASE_DIR = os.getcwd()
    host = "ubuntu"

class PacketsTypes(Enum):
    GPS = 0x02
    BATTERY_SENSOR = 0x08
    HEARTBEAT = 0x0B
    VIDEO_TRANSMITTER = 0x0F
    LINK_STATISTICS = 0x14
    RC_CHANNELS_PACKED = 0x16
    ATTITUDE = 0x1E
    DEVICE_INFO = 0x29

PAYLOADS_SIZE: "dict[PacketsTypes, int]" = {
    PacketsTypes.GPS: 15,
    PacketsTypes.BATTERY_SENSOR: 8,
    PacketsTypes.HEARTBEAT: 1,
    PacketsTypes.VIDEO_TRANSMITTER: 6,
    PacketsTypes.LINK_STATISTICS: 10,
    PacketsTypes.RC_CHANNELS_PACKED: 22,
    PacketsTypes.ATTITUDE: 6,
    PacketsTypes.DEVICE_INFO: 30,
}

class PacketValidationStatus(Enum):
    VALID = 1
    UNKNOWN = 2
    INVALID = -1
    CRC = -2

_crc_tab = [0x00,0xD5,0x7F,0xAA,0xFE,0x2B,0x81,0x54,0x29,0xFC,0x56,0x83,0xD7,0x02,0xA8,0x7D,0x52,0x87,0x2D,0xF8,0xAC,0x79,0xD3,0x06,0x7B,0xAE,0x04,0xD1,0x85,0x50,
			0xFA,0x2F,0xA4,0x71,0xDB,0x0E,0x5A,0x8F,0x25,0xF0,0x8D,0x58,0xF2,0x27,0x73,0xA6,0x0C,0xD9,0xF6,0x23,0x89,0x5C,0x08,0xDD,0x77,0xA2,0xDF,0x0A,0xA0,0x75,
			0x21,0xF4,0x5E,0x8B,0x9D,0x48,0xE2,0x37,0x63,0xB6,0x1C,0xC9,0xB4,0x61,0xCB,0x1E,0x4A,0x9F,0x35,0xE0,0xCF,0x1A,0xB0,0x65,0x31,0xE4,0x4E,0x9B,0xE6,0x33,
			0x99,0x4C,0x18,0xCD,0x67,0xB2,0x39,0xEC,0x46,0x93,0xC7,0x12,0xB8,0x6D,0x10,0xC5,0x6F,0xBA,0xEE,0x3B,0x91,0x44,0x6B,0xBE,0x14,0xC1,0x95,0x40,0xEA,0x3F,
			0x42,0x97,0x3D,0xE8,0xBC,0x69,0xC3,0x16,0xEF,0x3A,0x90,0x45,0x11,0xC4,0x6E,0xBB,0xC6,0x13,0xB9,0x6C,0x38,0xED,0x47,0x92,0xBD,0x68,0xC2,0x17,0x43,0x96,
			0x3C,0xE9,0x94,0x41,0xEB,0x3E,0x6A,0xBF,0x15,0xC0,0x4B,0x9E,0x34,0xE1,0xB5,0x60,0xCA,0x1F,0x62,0xB7,0x1D,0xC8,0x9C,0x49,0xE3,0x36,0x19,0xCC,0x66,0xB3,
			0xE7,0x32,0x98,0x4D,0x30,0xE5,0x4F,0x9A,0xCE,0x1B,0xB1,0x64,0x72,0xA7,0x0D,0xD8,0x8C,0x59,0xF3,0x26,0x5B,0x8E,0x24,0xF1,0xA5,0x70,0xDA,0x0F,0x20,0xF5,
			0x5F,0x8A,0xDE,0x0B,0xA1,0x74,0x09,0xDC,0x76,0xA3,0xF7,0x22,0x88,0x5D,0xD6,0x03,0xA9,0x7C,0x28,0xFD,0x57,0x82,0xFF,0x2A,0x80,0x55,0x01,0xD4,0x7E,0xAB,
			0x84,0x51,0xFB,0x2E,0x7A,0xAF,0x05,0xD0,0xAD,0x78,0xD2,0x07,0x53,0x86,0x2C,0xF9]

# Calculate CRC for CRSF packet 
def crsf_crc(data: bytes) -> int:
    """
    Calculate CRC for given data.    

    @data: Data to calculate CRC for
    @return: calculated CRC
    """
    crc = 0
    for i in data:
        crc = _crc_tab[crc ^ i]
    return crc

# Validate if an incoming packet matches expected size for its type 
def validate_packet(data) -> Tuple[PacketValidationStatus, int, PacketsTypes]:
	"""
    Validate that a given packet is the right size

    @data: Data that contains potential CRSF packet
    @return: Packet Validation status, size of packet, type of packet
    """
	size = data[0] - 2
	type: PacketsTypes = data[1]
	try:
		expected_size = PAYLOADS_SIZE.get(PacketsTypes(type), None)
	except:
		expected_size = None
	if not expected_size:
		# print(f"Packet not matching expected size! Expected: {expected_size}\tReceived:{size}")
		return [PacketValidationStatus.UNKNOWN, size, type]
	if expected_size == size or expected_size == FRSKY_RX_INFO_LEN or expected_size == BETAFPV_RX_INFO_LEN:
		# print(f"Packet matches expected size! Expected: {expected_size}\tReceived:{size}")
		return [PacketValidationStatus.VALID, size, type]
	# print(f"Packet invalid!")
	return [PacketValidationStatus.INVALID, size, type]

# Parse device info response
def parse_device_info_response(input: bytearray) -> Union[bytearray, None]:
	"""
	Parse the response of a request for device information

	@input: a bytearray representing the unfiltered data read from the UART
	@return: bytearray which contains the payload
							or 
					None if invalid or no packet		
	"""
	while len(input) > 0:
		val = input[0]
		# print(f"Val: {hex(val)}")
		if val == SYNC_BYTE:
			status, size, frame_type = validate_packet(input[1:])
			# print(f"Frame Type: {frame_type}")
			if status == PacketValidationStatus.VALID and PacketsTypes(frame_type) == PacketsTypes.DEVICE_INFO:
				# print(f"Input: {input}")
				payload = input[CRSF_HEADER_SIZE-3:CRSF_HEADER_SIZE + size] 
				if payload[-1] == SYNC_BYTE: payload = payload[:-1]
				# print("Payload, len(payload)")
				# print(payload, len(payload))
				packet_crc = payload[-1]
				crc = crsf_crc(payload[:-1])
				if crc != packet_crc:
					status = PacketValidationStatus.CRC
					# print(f"invalid CRC, expected:{crc}\tactual:{packet_crc}")
					return None
				else:
					# print(f"Valid Packet, CRC is as expected:{crc}\tactual:{packet_crc}")
					return payload[3:]
		del input[0]
	return None


uint8_t = ct.c_uint8
if host == "voxl2":
    dev = voxl_serial.VoxlSerialPort()
    dev.port = '/dev/tty-slpi-7'
    dev.baudrate = 420000
    dev.open()
else:
    # Automatically opens serial
    print("Using /dev/ttyUSB0")
    dev = serial.Serial("/dev/ttyUSB0",420000, timeout=1)

retries : int = 0
bytes_read : int = 0
device: str = ""
_device_name:str =""
_sw_version:str =""

for retry in range(5):
    data_out = (uint8_t * 7)(0xEC, 0x05, 0x28, 0xEC, 0xC8, 0xEC, 0x98)
    dev.write(bytearray(data_out))
    time.sleep(0.01)
    data = dev.read(64)
    response = parse_device_info_response(data)
    if response:
        # print(response)
        for byte in response:
            if byte == 0:
                _device_name = device
                break				
            device += chr(byte)
    
        # Get Version -> Version is always the [-7:-3] bytes in the packet we receive in the current version of ELRS
        _sw_version = '.'.join([str(int(byte)) for byte in response[-7:-3]])
        if _sw_version[0] == "0":
            _sw_version = _sw_version[2:]
        break
    retries += 1

if retries == 5:
    print(f"[ERROR] Error reading device information.")
    dev.close()
    exit(-1)

dev.close()
print(f"{_device_name}:{_sw_version}")
# return _device_name


# Device Info command
# data_out = (uint8_t * 7)()
# data_out[0] = 0xEC  # ADDR ----> RECEIVER
# data_out[1] = 0x05  # LEN  ----> 5 BYTES
# data_out[2] = 0x28  # TYPE ----> DEVICE PING
# data_out[3] = 0xEC  # DEST ----> RECEIVER 
# data_out[4] = 0xC8  # ORIG ----> FLIGHT CONTROLLER
# data_out[5] = 0xEC  # PAYLOAD -> RECEIVER
# data_out[6] = 0x98  # CRC -----> 152

# bytes_out = bytearray(data_out)
# data = (uint8_t * 32)()

# # Write command to UART port
# # print(f'\nSending {len(data_out)} bytes: {list(data_out)}')
# dev.write(bytes_out)

# time.sleep(0.25)

# RESPONSE FORMAT:
# DEV ADDR/SYNC BYTE                      -> 0xC8
# LEN/FRAME SIZE                          -> 0x20 -> 32
# TYPE: DEVICE_INFO                       -> 0x29
# DEST: CRSF_ADDRESS_FLIGHT_CONTROLLER    -> 0xC8
# ORIGIN: CRSF_ADDRESS_CRSF_RECEIVER      -> 0xEC
# Payload
# CRC

# Read response 
# data = dev.read(64)
# parse_response(data)
# dev.close()
# hex_string = ' '.join([format(byte, '02x') for byte in data])

# Print the number of bytes read and hex values
# print(len(data))
# print(hex_string)

# Extra prints msgs useful for debugging...
# print(f"Read {len(data)} bytes: {hex_string}")
# non_zero_values = [value for value in data if value != 0]
# print(f"Decimal bytes: {non_zero_values}")
# print(f"Hex bytes: {[hex(value) for value in non_zero_values]}")
# print(f"Decoded bytes: {[chr(value) for value in data if 0 <= value <= 128]}")
