import serial
import ctypes as ct
import argparse
import os

CRSF_CRC_POLY = 0xD5
CRCLEN = 256
FRAME_SIZE = 5
uint8_t = ct.c_uint8

# Detect if we are on a voxl2 or host computer
if os.path.exists('/usr/share/modalai/voxl-elrs'):
    import voxl_serial
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "voxl2"
else:
    BASE_DIR = os.getcwd()
    host = "ubuntu"

class GENERIC_CRC8:
    def __init__(self, poly= 0xD5):
        self.crcpoly = poly
        self.crc8tab = [0, 213, 127, 170, 254, 43, 129, 84, 41, 252, 86, 131, 215, 2, 168, 125, 82, 135,
                        45, 248, 172, 121, 211, 6, 123, 174, 4, 209, 133, 80, 250, 47, 164, 113, 219, 14, 
                        90, 143, 37, 240, 141, 88, 242, 39, 115, 166, 12, 217, 246, 35, 137, 92, 8, 221, 
                        119, 162, 223, 10, 160, 117, 33, 244, 94, 139, 157, 72, 226, 55, 99, 182, 28, 201, 
                        180, 97, 203, 30, 74, 159, 53, 224, 207, 26, 176, 101, 49, 228, 78, 155, 230, 51, 
                        153, 76, 24, 205, 103, 178, 57, 236, 70, 147, 199, 18, 184, 109, 16, 197, 111, 186, 
                        238, 59, 145, 68, 107, 190, 20, 193, 149, 64, 234, 63, 66, 151, 61, 232, 188, 105, 
                        195, 22, 239, 58, 144, 69, 17, 196, 110, 187, 198, 19, 185, 108, 56, 237, 71, 146, 
                        189, 104, 194, 23, 67, 150, 60, 233, 148, 65, 235, 62, 106, 191, 21, 192, 75, 158, 
                        52, 225, 181, 96, 202, 31, 98, 183, 29, 200, 156, 73, 227, 54, 25, 204, 102, 179, 231,
                        50, 152, 77, 48, 229, 79, 154, 206, 27, 177, 100, 114, 167, 13, 216, 140, 89, 243, 
                        38, 91, 142, 36, 241, 165, 112, 218, 15, 32, 245, 95, 138, 222, 11, 161, 116, 9, 220,
                        118, 163, 247, 34, 136, 93, 214, 3, 169, 124, 40, 253, 87, 130, 255, 42, 128, 85, 1, 
                        212, 126, 171, 132, 81, 251, 46, 122, 175, 5, 208, 173, 120, 210, 7, 83, 134, 44, 249 ]

    def calc(self, data, length , crc:uint8_t=0) -> uint8_t:
        for i in range(length):
            crc= self.crc8tab[crc ^ data[i]]
        return crc


def parse_args() -> (int, int):
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-c', '--channel', help='Telemetry input channel (from transmitter)', required=True, type=str)
    parser.add_argument('-p', '--pin', help='PWM output pin (on peripheral)', required=True, type=str)
    args = vars(parser.parse_args())
    
    try:
        input = int(args["channel"])
    except:
        print(f"[ERROR] Invalid Input Channel Value: {args['channel']}.")
        exit(-3)
    
    if (input < 1 or input > 16) and input != 255:
        print("[ERROR] Invalid Input channel value! Must be between 1-16. set_pwm_ch")
        exit(-2)

    try:
        pin = int(args["pin"])
    except:
        print(f"[ERROR] Invalid Output Pin Value: {args['pin']}.")
        exit(-3)
    
    if pin < 1 or pin > 4:
        print("[ERROR] Invalid Output Pin value! Must be between 1-4.")
        exit(-2)

    return input, pin

if host == "voxl2":
    import voxl_serial
    dev = voxl_serial.VoxlSerialPort()
    dev.port = '/dev/tty-slpi-7'
    dev.baudrate = 420000
    dev.open()
else:
    # Automatically opens serial
    print("Using /dev/ttyUSB0")
    dev = serial.Serial("/dev/ttyUSB0",420000)

input, pin = parse_args()
crc8 = GENERIC_CRC8(poly=CRSF_CRC_POLY)

# PWM command 
data_out = (uint8_t * (FRAME_SIZE + 2))()
data_out[0] = 0xEC      # ADDR    -> RECEIVER
data_out[1] = 0x05      # LEN     -> FRAME_SIZE bytes
data_out[2] = 0x32      # TYPE    -> COMMAND
data_out[3] = 0xF3      # CMD     -> MODALAI SET PWM CHANNEL
data_out[4] = pin       # PIN     -> pin #
data_out[5] = input     # INPUT   -> ch #
data_out[6] = crc8.calc(data_out[2:],FRAME_SIZE - 1, 0)  # CRC   -> Calculated value

bytes_out = bytearray(data_out)
print(f'\nSending {len(data_out)} bytes: {list(data_out)}')
hex_string = ' '.join([format(byte, '02x') for byte in bytes_out])
print(f"{hex_string.upper()}")
dev.write(bytes_out)
dev.close()
